/*
 *
 * Original Author: Luke Stras <stras@ecf.toronto.edu>
 *
 * Revision History
 * 1999/09/22 stras Initial version
 * 1999/09/25 JHS   re-use for PPPOE discovery 
 * 1999/09/26 JHS   retransmission for PADI and PADR
 * 1999/10/11 ireid@sympatico.ca   -I fix
 * 1999/10/22  jhs AC cookie echo, -F, -R, -P and help extensions
 * 1999/10/23  small fix by Josef Drexler to create_padr(); 
 *		doesnt make sense and doesnt work if removed 
 *		(will figure it out later). Looks like alignm issue
 *		Fixed: Stupid mistake on my part (much cleaner now).
 * 1999/10/27   Woody Suwalski <woody@rebel.com> ARM requires packed
 *		structs.
 * 1999/10/29	Andi Kleen <ak@suse.de> Port to new ioctl interface,
 *		Fix Endian bugs.
 *
 * 1999/11/20  jhs: -P option
 * 1999/11/30  Marc Boucher <marc@mbsi.ca> sigchild and misc cleanup, typos etc
 * 1999/12/04  jhs: AC and service name reflection + overriding (-A, -E)
 *		major rewrite of the PADR creation. Much cleaner now.
 * 1999/12/09  jhs: PADI override of Sname;
 * 1999/12/13  Marc Boucher <marc@mbsi.ca>: support for daemonization, syslog,
 *             runtime debugging; include cleanup, ARM fixes, etc.
 *	       
 */

#include <stdio.h>		/* stdio               */
#include <stdlib.h>		/* strtoul(), realloc() */
#include <unistd.h>		/* STDIN_FILENO,exec    */
#include <string.h>		/* memcpy()             */
#include <errno.h>		/* errno                */
#include <signal.h>
#include <getopt.h>
#include <stdarg.h>
#include <syslog.h>
#include <paths.h>

#include <sys/types.h>		/* socket types         */
#include <asm/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>		/* ioctl()              */
#include <sys/select.h>
#include <sys/socket.h>		/* socket()             */

#include <linux/sockios.h>	/* ifreq struct         */
#include <linux/if_ether.h>

#include <net/if.h>		/* ifreq struct         */
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>

#include "pppoed.h"

#include <linux/pppoe.h>
#if POE_VERSION != 1
#warning "User utils may be outdated"
#endif

/* Bigger than the biggest ethernet packet we'll ever see, in bytes */
#define MAX_PACKET	2000

int ctrl_fd;
int opt_debug = 0;
int opt_daemonize = 1;

char pppox[6];

/* 
   pointer to packet we have to retransmit
 */

struct pppoe_packet *ret_pkt;
int ret_pkt_size = 0;
int state = CODE_PADI;

/*
   retransmit retries for the PADR and PADI packets
   during discovery
 */
int PADR_ret;
int PADI_ret;


char *sname=NULL;  /* service name */
char *acname=NULL; /* AC name */
struct pppoe_tag *ctag = NULL;
struct pppoe_tag *stag = NULL;
struct pppoe_tag *ntag = NULL;


static struct session ses_table[16];

int disc_sock, sess_sock, socks[3];	/* raw socket we use */
char if_name[IFNAMSIZ];			/* interface to use */

#if 0
/* just for testing purposes ignore otherwise */
void
read_setup ()
{
/* should really populate the table with the params
   read from some config file; lets stick with one for now
 */
/* debugging for the 10.0.0.1 machine */
  unsigned char dstmac[6] = { 0x0, 0x0, 0xe8, 0x1c, 0x6, 0x9b };

  ses_table[0].pid = -1;
  ses_table[0].state = CONNECTED;
  ses_table[0].line = 0;
  ses_table[0].sid = 12;	/* session ID */
  strcpy (ses_table[0].name, "eth0");	/*name */
  memcpy (&ses_table[0].dmac, dstmac, 6);
}
#endif

int
create_raw_socket (unsigned short type)
{
  int optval = 1, rv;

  if ((rv = socket (PF_INET, SOCK_PACKET, htons (type))) < 0) {
    error ("socket: %m");
    return -1;
  }

  if (setsockopt (rv, SOL_SOCKET, SO_BROADCAST, &optval, sizeof (optval)) < 0) {
    error ("setsockopt: %m");
    return -1;
  }

  return rv;
}

int
get_hw_addr (int s, char *if_name, char *hw_addr)
{
  struct ifreq ifr;

  strncpy (ifr.ifr_name, if_name, sizeof (ifr.ifr_name));

  if (ioctl (s, SIOCGIFHWADDR, &ifr) < 0) {
    error ("ioctl(SIOCGIFHWADDR): %m");
    return -1;
  }

  if (ifr.ifr_hwaddr.sa_family != ARPHRD_ETHER) {
    error ("interface %s is not Ethernet!", if_name);
    return -1;
  }

  memcpy (hw_addr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);

/*
   dbglog("<%s> we received hware address %E",if_name,hw_addr);

 */
  return 0;
}

int
create_padi (struct pppoe_packet **packet, const char *src, const char *name)
{
  int size;

  if (packet == NULL)
    return 0;

  if (*packet != NULL)
    free (*packet);

  size = sizeof (struct pppoe_packet) + sizeof (struct pppoe_tag);
  if (stag)
	  size+=ntohs (stag->length);

  if ((*packet = malloc (size)) == NULL) {
    error ("create_padi: malloc failed: %m");
    return 0;
  }

  memcpy ((*packet)->ethhdr.h_dest, MAC_BCAST_ADDR, ETH_ALEN);
  memcpy ((*packet)->ethhdr.h_source, src, ETH_ALEN);
  (*packet)->ethhdr.h_proto = htons (ETH_P_PPPOE_DISC);
  (*packet)->ver = 1;
  (*packet)->type = 1;
  (*packet)->code = CODE_PADI;
  (*packet)->session = 0;
  (*packet)->length = htons (size - sizeof (struct pppoe_packet));

  if (!stag) {
  /* fill out a blank service-name tag */
	  (*(struct pppoe_tag *) (*packet + 1)).type = htons (TAG_SERVICE_NAME);
	  (*(struct pppoe_tag *) (*packet + 1)).length = 0;
   } else {
	  memcpy ((char *) (*packet + 1), stag, ntohs (stag->length));
   }
   

  return size;
}



int
create_padr (struct pppoe_packet **packet, const char *src, const char *dst)
{
  int size;
  int size1 = 0, size2 = 0,size3=0;
  char *p;

  if (packet == NULL)
    return 0;

  if (*packet != NULL)
    free (*packet);

  if (stag)
    size1 = sizeof (struct pppoe_tag) + ntohs (stag->length);
  if (ntag)
    size2  = sizeof (struct pppoe_tag) + ntohs (ntag->length);
  if (ctag)
    size3  = sizeof (struct pppoe_tag) + ntohs (ctag->length);

   size=size1+size2+size3;
  if ((*packet = malloc (size + sizeof (struct pppoe_packet))) == NULL) {
    error ("create_padr: malloc failed: %m");
    return 0;
  }

  memcpy ((*packet)->ethhdr.h_dest, dst, ETH_ALEN);
  memcpy ((*packet)->ethhdr.h_source, src, ETH_ALEN);
  (*packet)->ethhdr.h_proto = htons (ETH_P_PPPOE_DISC);
  (*packet)->ver = 1;
  (*packet)->type = 1;
  (*packet)->code = CODE_PADR;
  (*packet)->session = 0;
  (*packet)->length = htons (size);

  p = (char *) (*packet + 1);
  if (stag) {
    memcpy ((char *) p, stag, size1);
  }
  p += size1;

 if (ntag) {
   memcpy ((char *) p, ntag, size2);
 }

p += size2;

if (ctag) {
   memcpy ((char *) p, ctag, size3);
}

return size + sizeof (struct pppoe_packet);
}


int
send_packet (int sock, struct pppoe_packet *packet, int len, const char *ifn)
{
  struct sockaddr addr;
  int c;

  memset (&addr, 0, sizeof (addr));
  dbglog ("send_packet: ifn %s %P", ifn, packet, len);
  strcpy (addr.sa_data, ifn);

  if ((c = sendto (sock, packet, len, 0, &addr, sizeof (addr))) < 0)
    error ("sendto (send_packet): %m");

  return c;
}

int
wait_for_packet (int *sock, int ns, struct pppoe_packet **packet, int *len,
		 int num_ret)
{
  fd_set fdset;
  int i, maxs = -1;
  struct sockaddr_in from;
  socklen_t fromlen;
  struct timeval timeout;
  int retransmits = 0;
  int reason = -1;

/* 
   a num_ret == 0 implies block and wait for the packet
 */

  FD_ZERO (&fdset);

  while (1) {
    retransmits++;
    if (num_ret && retransmits > num_ret)
      return -1;
    if (retransmits <= num_ret) {
      timeout.tv_sec = 1 << retransmits;
      timeout.tv_usec = 0;
    }

    for (i = 0; i < ns; i++) {
      FD_SET (sock[i], &fdset);
      if (sock[i] > maxs)
	maxs = sock[i];
    }

    if (num_ret == 0)
      reason = select (maxs + 1, &fdset, NULL, NULL, NULL);
    else
      reason = select (maxs + 1, &fdset, NULL, NULL, &timeout);

    if (reason < 0) {
      error ("select (wait_for_packet): %m");
      return -1;
    }
    if (reason == 0) {		/* timeout */
      if (send_packet (*sock, ret_pkt, ret_pkt_size, if_name) < 0) {
	warn ("unable to send PADI packet");
	return -1;
      }
      continue;			/* go back to the top */
    }

    /* reason has the number of ready descriptors */


    for (i = 0; i < ns; i++) {
      if (FD_ISSET (sock[i], &fdset)) {
	if (sock[i] < 3)	/* ie, it's an fd */
	  return sock[i];

	if (*packet != NULL)
	  free (*packet);

	if ((*packet = malloc (MAX_PACKET)) == NULL) {
	  error ("wait_for_packet: malloc failed: %m");
	  return -1;
	}

	fromlen = sizeof (from);
	if (recvfrom (sock[i], *packet, MAX_PACKET, 0,
		      (struct sockaddr *) &from, &fromlen) < 0) {
	  error ("wait_for_packet: recv failed: %m");
	  return -1;
	}


	return sock[i];
      }
    }
  }
}


struct pppoe_tag *
scan_tag (int length, struct pppoe_tag *start_tag, unsigned short type)
{
  struct pppoe_tag *tag, *ret;

  tag = start_tag;
  while ((tag < (start_tag + length)) && (ntohs (tag->type) != 0)) {
    if(DEB_DISC2) {
      dbglog ("scan_tag: tag %p length %d tag->type %x data: %.*B", tag,
	      ntohs (tag->length), ntohs (tag->type), ntohs (tag->length), (char *) (tag + 1));
    }
    if (ntohs (tag->type) == type) {	/* got it */

      if(DEB_DISC2)
        dbglog ("Found it!");
      if ((ret = malloc (sizeof (struct pppoe_tag) + tag->length)) == NULL) {
	error ("scan_tag: malloc failed: %m");
	die (-1);
      }

      memcpy (ret, tag, sizeof (struct pppoe_tag) + ntohs (tag->length));

      return ret;
    }
    if(DEB_DISC2)
      dbglog ("scan_tag: not matched! ");

    tag = (struct pppoe_tag *)((char *) (tag) + sizeof (struct pppoe_tag) + ntohs (tag->length));

  }

  return NULL;
}

int
server_discover (int lineid)
{

  char src_addr[ETH_ALEN];	/* source hardware address */
  char dst_addr[ETH_ALEN];	/* destination hardware address */

  int pkt_size;

  struct pppoe_packet *packet = NULL;

  int state = CODE_PADI;



  /* main discovery loop */

  if (get_hw_addr (disc_sock, if_name, src_addr) != 0) {
    error ("unable to get hardware address");
    return -1;
  }


  while (1) {
    if (state == STATE_RUN)
      break;

    switch (state) {
    case CODE_PADI:		/* initiate ppp_connection */
      if(DEB_DISC)
       dbglog ("Sending PADI");
      /* start the PPPoE session */
      if ((pkt_size = create_padi (&packet, src_addr, NULL)) == 0) {
	error ("unable to create PADI packet");
	die (1);
      }
      /* send the PADI packet */
      if (send_packet (disc_sock, packet, pkt_size, if_name) < 0) {
	warn ("unable to send PADI packet");
	die (1);
      }

      ret_pkt = packet;
      ret_pkt_size = pkt_size;

      state = CODE_PADO;
      break;

    case CODE_PADO:		/* wait for PADO */
      if(DEB_DISC)
        dbglog ("waiting for PADO");
      if (wait_for_packet (&disc_sock, 1, &packet, &pkt_size, PADI_ret)
	  != disc_sock || (packet->code != CODE_PADO &&
			   packet->code != CODE_PADT)) {
	if (packet->code == CODE_PADI)	/* just an echo */
	  continue;
	warn ("unexpected packet %x", packet->code);
	continue;
      }
      if (packet->code == CODE_PADT) {	/* early termination */
	state = CODE_PADT;
	continue;
      }
      if(DEB_DISC2)
        dbglog ("PADO received: %P", packet, pkt_size);
      memcpy (dst_addr, packet->ethhdr.h_source, ETH_ALEN);
/* some really nasty coding here; try to re-use M Ostrowskis code
instead at some point
*/
      if (ntohs (packet->length) > 0) {
    if (DEB_DISC2 && !stag)
		dbglog ("Scanning for Service Name ");
    if (!stag)
	stag=
	  scan_tag (ntohs (packet->length),
		    (struct pppoe_tag *) (packet + 1), TAG_SERVICE_NAME);

    if (DEB_DISC2 && !ntag)
		dbglog ("Scanning for AC Name ");
    if (!ntag)
	ntag=
	  scan_tag (ntohs (packet->length),
		    (struct pppoe_tag *) (packet + 1), TAG_AC_NAME);

    if (DEB_DISC2 && !ctag)
		dbglog ("Scanning for AC cookies");
    if (!ctag)
	ctag =
	  scan_tag (ntohs (packet->length),
		    (struct pppoe_tag *) (packet + 1), TAG_AC_COOKIE);
	}

      state = CODE_PADR;
      break;

    case CODE_PADR:		/* send PADR */
      if(DEB_DISC)
        dbglog ("Sending PADR");
      if ((pkt_size = create_padr (&packet, src_addr, dst_addr)) == 0) {
	error ("unable to create PADR packet");
	die (1);
      }

/*
probably free all the tags sent here; who cares really ... when
the prog exists the system would reclaim the memory anyway ..
*/
      if (send_packet (disc_sock, packet, pkt_size, if_name) < 0) {
	error ("unable to send PADR packet");
	die (1);
      }
      state = CODE_PADS;
      ret_pkt = packet;
      ret_pkt_size = pkt_size;
      break;

    case CODE_PADS:		/* wait for PADS */
      if(DEB_DISC)
        dbglog ("Waiting for PADS");
      if (wait_for_packet (&disc_sock, 1, &packet, &pkt_size, PADR_ret)
	  != disc_sock || (packet->code != CODE_PADS &&
			   packet->code != CODE_PADT)) {
	if (packet->code == CODE_PADR)	/* just an echo */
	  continue; 
	warn ("unexpected packet %x", packet->code);
	continue;
      }
      if (memcmp (packet->ethhdr.h_source, dst_addr, sizeof (dst_addr))
	  != 0)
	continue;		/* discard packets not from AC */
      if (packet->code == CODE_PADT) {	/* early termination */
	state = CODE_PADT;
	continue;
      }

      socks[0] = disc_sock;

      ses_table[lineid].line = lineid;
      ses_table[lineid].sid = packet->session;
      memcpy (&ses_table[lineid].dmac, dst_addr, ETH_ALEN);

      notice ("Discovery completed successfully (session ID=%d [0x%x], AC MAC=%E)",
		ntohs (ses_table[lineid].sid), ntohs (ses_table[lineid].sid),
		ses_table[lineid].dmac);

      state = STATE_RUN;

/* should start a thread here which just waits for a PADT
   packet 
 */
      break;

    case STATE_RUN:		/* running */

      break;

    case CODE_PADT:		/* received a PADT */
      return -1;
      break;

    default:
      error ("invalid state %x", state);
      die (1);
    }
  }
  return (0);
}

void
sigproc (int src)
{
  int i;
  info ("Received signal %d", src);
  for (i = 0; i < 16; i++) {
    if (ses_table[i].pid > 0) {
      dbglog ("Sending signal %d to pid %d",src,ses_table[i].pid);
      kill(ses_table[i].pid,src);
    }
  }
  die (1);
}



void
sigchild (int src)
{
  pid_t pid;
  int status;
  int i;
  pid = waitpid (-1, &status, WNOHANG);
  info ("Child received signal %d PID %d, status %d", src, pid, status);
  if (pid < 1) {
    return;
  }
/* scan */
  for (i = 0; i < 16; i++) {
    if (ses_table[i].pid == pid) {
      info ("Line id which died is: %d", i);
      ses_table[i].state = DISCONNECTED;
      break;
    }

  }

}


int
open_device (void)
{
	int fd;
	char fn[30]; 
	sprintf(fn,"/dev/%s",pppox);
	fd = open(fn,O_RDWR);
	if (fd < 0 ){ 
		error ("cannot open %s: %m", fn);
		die(1);
	}
	return fd;
}

#define err(x)  error(x),die(1)

int inet_sk = -1; 

void open_inet_sk(void)
{
	if (inet_sk != -1)
		return;
	inet_sk = socket(PF_INET,SOCK_DGRAM,0);
	if (inet_sk<0) err("inet socket: %m"); 
}

int ifindex(char *name)
{
	struct ifreq ifr;
	memset(&ifr,0,sizeof ifr); 
	strncpy(ifr.ifr_name, name, sizeof(ifr.ifr_name)); 
	if (ioctl(inet_sk, SIOCGIFINDEX, &ifr) < 0) {
		error ("error getting device %s: %m", name);
		die(1);
	}
	return ifr.ifr_ifindex;
}

int 
cntrl (int lineid)
{
	struct poe_arg poe;

	open_inet_sk();

	memset(&poe,0,sizeof poe); 
	poe.poe_version = POE_VERSION; 

	/* Set session ID */ 
	dbglog ("cntrl: SessionID (%x)",ses_table[lineid].sid); 
	/* sid is already in network order; we didnt bother converting 
	earlier */
	poe.poe_session_id = ses_table[lineid].sid; 
	if (ioctl(ctrl_fd, PPPOEIOCSID, &poe) < 0) err("PPPOECSID: %m"); 

	/* Set device */ 
	poe.poe_addr.sll_ifindex =  ifindex( ses_table[lineid].name );
	poe.poe_addr.sll_hatype = ARPHRD_ETHER;
	poe.poe_addr.sll_halen = 6;
	poe.poe_addr.sll_family = 0; /* XXX */
	/* should set protocol */ 
	if (ioctl(ctrl_fd, PPPOEIOCSDEV, &poe) < 0) err("PPPOECSDEV: %m"); 

	/* Destination MAC */
	memcpy(poe.poe_addr.sll_addr, ses_table[lineid].dmac, 6); 	
	if (ioctl(ctrl_fd, PPPOEIOCDMAC, &poe) < 0) err("PPPOEIODID: %m"); 

	return 0;
}
	
	
/* **********************************************88
*/
int lineid = 0;
char *fname = NULL;
char *pppd = _PATH_PPPD;
int num_restart = 1;		/* 1 means no restart */

int
ppp_connect (void)
{

  int pid;

restart:
  num_restart--;

  info ("ppp_connect: total retries to go %d", num_restart);

  if ((disc_sock = create_raw_socket (ETH_P_PPPOE_DISC)) < 0) {
    error ("unable to create raw socket");
    return 1;
  }


/* we block until we extablish a ppp_connection
   in this version
 */

  if (server_discover (lineid) < 0) {
    if (!num_restart)
	die(-1);
    goto restart;
  }

  close (disc_sock);		/* dont need it anymore */
/* at this point we know the destination MAC address,
   the session ID,
   the device,
 */

#if 0
/* for debugging only, turn this off in the real thing */
/* populate table */

  read_setup ();
#endif
  ctrl_fd = open_device ();

/* download the cntrl info for this lineid */
  cntrl (lineid);

  close (ctrl_fd);

  signal (SIGCHLD, &sigchild);

/* 
   now lets fire the PPP session
   TODO: use pthreads 
 */

  pid = fork ();
  if (pid < 0) {
    error ("unable to fork() for pppd: %m");
    die (-1);
  }

  ses_table[lineid].pid = pid;
  if (!pid) {
    /* the speed is meaningless really, but pppd insists */
    if(fname) {
      info ("starting %s to interface %s using options file: %s",
	      pppd, fname, pppox);

      if (execl (pppd, "pppd", pppox, "38400", "file", fname, NULL) < 0) {
        error ("cannot execl %s: %m", pppd);
        die (-1);
      }
    } else {
      info ("starting %s to interface %s", pppd, pppox);

      if (execl (pppd, "pppd", pppox, "38400", NULL) < 0) {
        error ("cannot execl %s: %m", pppd);
        die (-1);
      }
    }
  }

  dbglog ("PPP started pid is: %d", ses_table[lineid].pid);

/* 
   sit here and asynchronously wait for a PADT
   or issue a PADT should the user decide to terminate the ppp_connection
   Hmmmm ... maybe we should have an option to restart the ppp_connection
   if it dies
 */

/* move to ppp_connected state only after pppd has been started */
  ses_table[lineid].state = CONNECTED;

  pause ();

  ses_table[lineid].state = DISCONNECTED;
  info ("OK we got killed");

  if (!num_restart)
    die (0);

  goto restart;
}

void
print_help ()
{
  notice ("pppoe version %d.%d build %d", VERSION_MAJOR, VERSION_MINOR,
	  VERSION_DATE);
  notice ("recognized options are:");
  notice (" -I <interface> : overrides the default interface of eth0");
  notice (" -L <pppox_num> : overrides the default lineid of pppox0. Used to");
  notice ("                  have multiple sessions");
  notice (" -S : starts pppoed in server mode");
  notice (" -R <num_retries>: forces pppoed to be restarted num_retries");
  notice ("                   should the other end be detected to be dead.");
  notice ("                   Needs lcp_echo. Read the INSTALL file instructions");
  notice (" -F <filename> : specifies additional ppp options file");
  notice (" -d <level> : sets debug level");
  notice (" -D : prevents pppoed from detaching itself and running in the background");
  notice (" -P <path to pppd> : selects a different pppd. Defaults to "_PATH_PPPD);
  notice (" -A <AC name> to select a specific AC by name");
  notice (" -E <AC service name> to select a specific AC service by name");
}


int
main (int argc, char **argv)
{

  int opt;
/* only worry about a single ppp_connection (ppp0) in this version
 */
/* the number of PADI and PADR retransmits.
   We should be able to use the command line options to pass these instead
 */
  PADR_ret = 5;
  PADI_ret = 5;

  log_to_fd = fileno(stdout);

/* defaults to eth0 */
  strcpy (if_name, "eth0");

  /* parse options */

  while ((opt = getopt (argc, argv, "A:E:d:DR:I:F:L:V:P:S:N")) != -1)
    switch (opt) {
    case 'R':			/* sets number of retries */
      num_restart = strtol (optarg, (char **) NULL, 10);
      num_restart += 1;
      break;
    case 'I':			/* sets interface */
      if (strlen (optarg) >= sizeof(if_name)) {
	error ("interface name cannot exceed %d characters", sizeof(if_name)-1);
	die (-1);
      }
      strncpy (if_name, optarg, sizeof(if_name));
      break;
    case 'F':			/* sets the options file */
      fname = optarg;
      break;
    case 'D':			/* don't daemonize */
      opt_daemonize = 0;
      break;
    case 'd':			/* debug level */
      opt_debug = strtol (optarg, (char **) NULL, 10);
      break;
    case 'P':			/* sets the pppd binary */
      pppd = optarg;
      break;
    case 'L':			/* lineid */
      lineid = strtol (optarg, (char **) NULL, 10);
      info ("lineid set to %s %d", optarg, lineid);
      break;
    case 'V':			/* version */
      notice ("pppoe version %d.%d build %d", VERSION_MAJOR,
	      VERSION_MINOR, VERSION_DATE);
      die (0);
    case 'S':			/* server mode */
      error ("pppoe server mode not supported in this version");
      die (0);
    case 'A':			/* server mode */
      info ("AC name override to %s", optarg);
      if ((ntag = malloc (sizeof (struct pppoe_tag) + strlen(optarg))) 
        == NULL) {
	error ("failed to malloc for AC name: %m");
	die(-1);
      }
	ntag->length=htons(strlen(optarg));
	ntag->type=htons(TAG_AC_NAME);
	strncpy ((char *)ntag+sizeof(struct pppoe_tag), optarg , strlen(optarg));
      break;
    case 'E':			/* server mode */
      info ("AC service name override to %s", optarg);
      /*jamal: malloc for the stag copy optarg as the value */
      if ((stag = malloc (sizeof (struct pppoe_tag) + strlen(optarg))) 
        == NULL) {
	error ("failed to malloc for service name: %m");
	die(-1);
      }
	stag->length=htons(strlen(optarg));
	stag->type=htons(TAG_SERVICE_NAME);
	strncpy ((char *)(ntag+sizeof(struct pppoe_tag)),optarg,strlen(optarg));
      break;
    default:
      error ("Unknown option '%c'", optopt);
      print_help ();
      die (1);
    }

  strcpy (ses_table[lineid].name, if_name);

/* careful with overflows here; this runs as root at 
the moment, so OK */
  sprintf (pppox, "pppox%d", lineid);

  create_pidfile ();

  openlog("pppoed", LOG_PID | LOG_NDELAY, LOG_PPPOE);
  setlogmask(LOG_UPTO( opt_debug ? LOG_DEBUG : LOG_INFO));

  if(opt_daemonize)
	detach();

  signal (SIGINT, &sigproc);
  signal (SIGTERM, &sigproc);

  ppp_connect ();

  die (0);
}
